
############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import os
from collections import namedtuple

import pytest

from src.polya_finder import PolyAFinder
from src.cage_finder import CagePeakFinder


PolyAAlignment = namedtuple('PolyAAlignment', ('query_name', 'cigartuples', 'seq', 'reference_start', 'reference_end'))
CageAlignment = namedtuple('CageAlignment', ('query_name', 'reference_name', 'reference_start',
                                             'reference_end', 'is_reverse'))


TUPLES1 = [(4, 2), (0, 15), (1, 1), (0, 5), (2, 1), (0, 16), (3, 4642), (0, 5), (1, 1), (0, 1), (1, 1),
           (0, 15), (2, 1), (0, 4), (1, 1), (0, 6), (2, 1), (0, 19), (1, 1), (0, 32), (1, 1), (0, 3),
           (3, 7649), (0, 25), (1, 1), (0, 13), (1, 1), (0, 29)]

TUPLES2 = [(0, 167), (4, 30)]

TUPLES3 = [(0, 167), (4, 30), (5, 5)]

TUPLES4 = [(0, 150), (3, 100), (0, 10), (3, 100), (0, 11), (4, 27)]

TUPLES5 = [(0, 147), (3, 100), (0, 5), (2, 2), (0, 4), (3, 100), (0, 5), (1, 2), (0, 5), (4, 30)]

TUPLES6 = [(0, 147), (3, 100), (0, 5), (2, 2), (0, 4), (3, 100), (0, 5), (1, 2), (0, 5), (4, 30), (5, 5)]

SEQ1 = 'CTCAAGACCAAGAAGGACGACATGACCATGGCTTAAAAGAGTCTGCTCCCCACAGCCCCCTGCGAT' \
       'GGATGGACGGAGGAACCAGGGTCGGACGACCTCCGATGCTAAGAGCACTCCAACTGCTGCAAACCG' \
       'AAGAAGCAGGCATCGGAGACACTCCCAAACCAGGAGCGACCAAGCACTGGCGCATGTGACTCAAG'

SEQ2 = 'CTCAAGACCAAGAAGGACGACATGACCATGGCTTAAAAGAGTCTGCTCCCCACAGCCCCCTGCGAT' \
       'GGATGGACGGAGGAACCAGGGTCGGACGACCTCCGATGCTAAGAGCACTCCAACTGCTGCAAACCG' \
       'AAGAAGCAGGCATCGGACDCCTTTTTTTTTTAGGCGCCAAAAAAAAAAAACCAAAAAAAAAAAAA'

SEQ3 = 'TTTTTTTTTTTTTTTTTTTTCTCAAGACCAAGAAGGACGACATGACCATGGCTTAAAAGAGTCTGC' \
       'TCCCCACAGCCCCCTGCGATGGATGGACGGAGGAACCAGGGTCGGACGACCTCCGATGCTAAGAGC' \
       'ACTCCAACTGCTGCAAACCGAAGAAGCAGGCATCGGACGCCAAAAAAAAACCAAAAAAAAAAAAA'

SEQ4 = 'TTTTTTTATTTGTTTTTTTTTTTTTTTCTCAAGACCAAGAAGGACGACATGACCATGGCTTAAAAGAGTCTGC' \
       'TCCCCACAGCCCCCTGCGATGGATGGACGGAGGAACCAGGGTCGGACGACCTCCGATGCTAAGAGC' \
       'ACTCCAACTGCTGCAAACCGAAGAAGCAGGCATCGGACGCCAAAAAAAAACCAAAAAAAAAAAAA'


class TestPolyAFinder:
    polya_finder = PolyAFinder()

    @pytest.mark.parametrize("alignment",
                             [PolyAAlignment('aligned_segment1', TUPLES1, SEQ1, 51054, 63535)])
    def test_no_t_head(self, alignment):
        assert self.polya_finder.find_polyt_external(alignment) == -1

    @pytest.mark.parametrize("alignment",
                             [PolyAAlignment('aligned_segment1', TUPLES1, SEQ1, 51054, 63535)])
    def test_no_a_tail(self, alignment):
        assert self.polya_finder.find_polya_external(alignment) == -1

    @pytest.mark.parametrize("alignment, expected",
                             [(PolyAAlignment('aligned_segment1', TUPLES2, SEQ2, 51054, 63535), 63538),
                              (PolyAAlignment('aligned_segment1', TUPLES3, SEQ2, 51054, 63535), 63538)])
    def test_find_a_tail(self, alignment, expected):
        assert self.polya_finder.find_polya_external(alignment) == expected

    @pytest.mark.parametrize("alignment, expected",
                             [(PolyAAlignment('aligned_segment1', TUPLES4, SEQ2, 1000, 1300), 1300),
                              (PolyAAlignment('aligned_segment2', TUPLES5, SEQ2, 1000, 1300), 1303)])
    def test_find_fake_a_tail(self, alignment, expected):
        assert self.polya_finder.find_polya_external(alignment) == expected

    @pytest.mark.parametrize("alignment",
                             [PolyAAlignment('aligned_segment1', list(reversed(TUPLES2)), SEQ3, 51054, 63535),
                              PolyAAlignment('aligned_segment1', list(reversed(TUPLES3)), SEQ3, 51054, 63535)])
    def test_find_t_head(self, alignment):
        assert self.polya_finder.find_polyt_external(alignment) == 51043

    @pytest.mark.parametrize("alignment, expected",
                             [(PolyAAlignment('aligned_segment1', list(reversed(TUPLES5)), SEQ4, 1000, 1300), 996),
                              (PolyAAlignment('aligned_segment2', list(reversed(TUPLES6)), SEQ4, 1000, 1300), 996)])
    def test_find_fake_t_head(self, alignment, expected):
        assert self.polya_finder.find_polyt_external(alignment) == expected

    @pytest.mark.parametrize("seq, expected",
                             [('TTTATATTGGTATTTTGCC', -1),
                              ('AAAAAAAAAAACGAAAAAAAAAAAAAAAA', 0),
                              ('TGGCGGTACAAAAAAAAAAACGAAAAAAAAAAAAAAAA', 9)])
    def test_find_polya_pos(self, seq, expected):
        assert self.polya_finder.find_polya(seq) == expected


class TestCagePeakFinder:
    source_dir = os.path.dirname(os.path.realpath(__file__))
    cage_peak_finder = CagePeakFinder(os.path.join(source_dir, 'toy_data/MAPT.Mouse.CAGE.bed'), 50)

    @pytest.mark.parametrize("alignment",
                             [CageAlignment('aligned_segment1', 'chr11', 24, 2165, False),
                              CageAlignment('aligned_segment2', 'chr11', 0, 1296, False),
                              CageAlignment('aligned_segment3', 'chr11', 80, 372, True)])
    def test_no_intersection(self, alignment):
        assert self.cage_peak_finder.find_cage_peak(alignment) == []

    @pytest.mark.parametrize("alignment",
                             [CageAlignment('aligned_segment1', 'chr11', 57, 522, False),
                              CageAlignment('aligned_segment2', 'chr11', 69, 1499, False)])
    def test_one_intersection(self, alignment):
        assert len(self.cage_peak_finder.find_cage_peak(alignment)) == 1

    @pytest.mark.parametrize("alignment",
                             [CageAlignment('aligned_segment1', 'chr11', 134, 1322, False),
                              CageAlignment('aligned_segment2', 'chr11', 102, 2191, False),
                              CageAlignment('aligned_segment2', 'chr11', 11, 126, True)])
    def test_several_hits(self, alignment):
        assert len(self.cage_peak_finder.find_cage_peak(alignment)) > 1




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from collections import defaultdict

import gffutils
from pyfaidx import Fasta, UnsupportedCompressionFormat

from .serialization import *
from .file_naming import *
from .isoform_assignment import BasicReadAssignment, ReadAssignmentType
from .assignment_io import (
    NormalTmpFileAssignmentLoader,
    QuickTmpFileAssignmentLoader, GeneListTmpFileAssignmentLoader
)
from .gene_info import GeneList, GeneInfo

logger = logging.getLogger('IsoQuant')


class BasicAssignmentLoader:
    def __init__(self, save_file_name):
        logger.info("Loading read assignments from " + save_file_name)
        assert os.path.exists(save_file_name)
        self.save_file_name = save_file_name

    def has_next(self):
        raise NotImplementedError()

    def get_next(self):
        raise NotImplementedError()


class FullAssignmentLoader(BasicAssignmentLoader):
    def __init__(self, save_file_name, multimapped_chr_dict, filtered_read_set=None):
        BasicAssignmentLoader.__init__(self, save_file_name)
        self.multimapped_chr_dict = multimapped_chr_dict
        self.filtered_read_set = filtered_read_set

    def load_read_assignment(self, read_assignment):
        if self.filtered_read_set is not None and read_assignment.read_id not in self.filtered_read_set:
            return None
        if self.multimapped_chr_dict is not None and read_assignment.read_id in self.multimapped_chr_dict:
            resolved_assignment = None
            for a in self.multimapped_chr_dict[read_assignment.read_id]:
                if a.assignment_id == read_assignment.assignment_id and a.chr_id == read_assignment.chr_id:
                    if resolved_assignment is not None:
                        logger.info("Duplicate read: %s %s %s" % (read_assignment.read_id, a.gene_id, a.chr_id))
                    resolved_assignment = a

            if not resolved_assignment:
                logger.warning("Incomplete information on read %s" % read_assignment.read_id)
                return None
            elif resolved_assignment.assignment_type == ReadAssignmentType.suspended:
                return None
            else:
                read_assignment.assignment_type = resolved_assignment.assignment_type
                read_assignment.gene_assignment_type = resolved_assignment.gene_assignment_type
                read_assignment.multimapper = resolved_assignment.multimapper
        return read_assignment


class ReadAssignmentLoader(FullAssignmentLoader):
    def __init__(self, save_file_name, gffutils_db, chr_record, multimapped_chr_dict, string_pools, filtered_read_set=None):
        FullAssignmentLoader.__init__(self, save_file_name, multimapped_chr_dict, filtered_read_set)
        self.genedb = gffutils_db
        self.chr_record = chr_record
        self.unpickler = NormalTmpFileAssignmentLoader(save_file_name, gffutils_db, chr_record, string_pools)

    def has_next(self):
        return self.unpickler.has_next()

    def get_next(self):
        if not self.unpickler.has_next():
            return None, None

        assert self.unpickler.is_gene_info()
        gene_info = self.unpickler.get_object()
        assignment_storage = []
        while self.unpickler.is_read_assignment():
            read_assignment = self.load_read_assignment(self.unpickler.get_object())
            if read_assignment is not None:
                assignment_storage.append(read_assignment)

        return gene_info, assignment_storage


class MergingSimpleReadAssignmentLoader(FullAssignmentLoader):
    def __init__(self, save_file_name, multimapped_chr_dict, string_pools, filtered_read_set=None):
        FullAssignmentLoader.__init__(self, save_file_name, multimapped_chr_dict, filtered_read_set)
        self.unpickler = GeneListTmpFileAssignmentLoader(save_file_name, string_pools)
        self.current_gene_list = None

    def has_next(self):
        return self.unpickler.has_next()

    def get_next(self):
        if not self.unpickler.has_next():
            return None, None

        assignment_storage = []
        while self.unpickler.has_next():
            if self.current_gene_list is None:
                assert self.unpickler.is_gene_info()
                self.current_gene_list = self.unpickler.get_object()
            elif self.unpickler.is_gene_info():
                gene_list = self.unpickler.get_object()
                if self.current_gene_list.overlaps(gene_list):
                    self.current_gene_list.merge(gene_list)
                else:
                    self.current_gene_list = gene_list
                    return None, assignment_storage

            while self.unpickler.is_read_assignment():
                read_assignment = self.load_read_assignment(self.unpickler.get_object())
                if read_assignment is not None:
                    assignment_storage.append(read_assignment)

        return None, assignment_storage


class MergingReadAssignmentLoader(MergingSimpleReadAssignmentLoader):
    def __init__(self, save_file_name, gffutils_db, chr_record, multimapped_chr_dict, filtered_read_set=None):
        MergingSimpleReadAssignmentLoader.__init__(save_file_name, multimapped_chr_dict, filtered_read_set)
        self.genedb = gffutils_db
        self.chr_record = chr_record

    def _create_gene_info(self):
        if not self.current_gene_list.gene_id_set:
            gene_info = GeneInfo.from_region(self.current_gene_list.chr_id, self.current_gene_list.start,
                                             self.current_gene_list.end, self.current_gene_list.delta)
        else:
            gene_info = GeneInfo([self.genedb[gene_id] for gene_id in self.current_gene_list.gene_id_set], self.genedb,
                                 self.current_gene_list.delta)
        if self.chr_record:
            gene_info.set_reference_sequence(gene_info.all_read_region_start,
                                             gene_info.all_read_region_end,
                                             self.chr_record)
        return gene_info

    def get_next(self):
        if not self.unpickler.has_next():
            return None, None

        assignment_storage = []
        while self.unpickler.has_next():
            if self.current_gene_list is None:
                assert self.unpickler.is_gene_info()
                self.current_gene_list = self.unpickler.get_object()
            elif self.unpickler.is_gene_info():
                gene_list = self.unpickler.get_object()
                if self.current_gene_list.overlaps(gene_list):
                    self.current_gene_list.merge(gene_list)
                else:
                    gene_info = self._create_gene_info()
                    for a in assignment_storage:
                        a.gene_info = gene_info
                    self.current_gene_list = gene_list
                    return gene_info, assignment_storage

            while self.unpickler.is_read_assignment():
                read_assignment = self.load_read_assignment(self.unpickler.get_object())
                if read_assignment is not None:
                    assignment_storage.append(read_assignment)

        gene_info = self._create_gene_info()
        for a in assignment_storage:
            a.gene_info = gene_info
        return gene_info, assignment_storage


class BasicReadAssignmentLoader(BasicAssignmentLoader):
    def __init__(self, save_file_name, string_pools):
        BasicAssignmentLoader.__init__(self, save_file_name)
        self.unpickler = QuickTmpFileAssignmentLoader(save_file_name, string_pools)

    def has_next(self):
        return self.unpickler.has_next()

    def get_next(self):
        if not self.unpickler.has_next():
            return

        assert self.unpickler.is_gene_info()
        self.unpickler.get_object()

        while self.unpickler.is_read_assignment():
            yield self.unpickler.get_object()


def prepare_multimapped_reads(saves_prefix ,chr_id, string_pools):
    multimapped_reads = defaultdict(list)
    multimap_loader = open(multimappers_file_name(saves_prefix ,chr_id), "rb")
    list_size = read_int(multimap_loader)
    while list_size != TERMINATION_INT:
        for i in range(list_size):
            a = BasicReadAssignment.deserialize(multimap_loader, string_pools)
            if a.chr_id == chr_id:
                multimapped_reads[a.read_id].append(a)
        list_size = read_int(multimap_loader)
    return multimapped_reads


def prepare_read_filter(chr_id, saves_prefix, use_filtered_reads):
    if not use_filtered_reads:
        return None
    filtered_reads = set()
    for l in open(filtered_reads_file_name(saves_prefix, chr_id), "r"):
        filtered_reads.add(l.rstrip())
    return filtered_reads


def load_genedb(genedb):
    if genedb:
        return gffutils.FeatureDB(genedb)
    return None


def create_assignment_loader(chr_id, saves_prefix, genedb, reference_fasta, reference_fai, string_pools, use_filtered_reads=False):
    current_chr_record = Fasta(reference_fasta, indexname=reference_fai)[chr_id]
    multimapped_reads = prepare_multimapped_reads(saves_prefix, chr_id, string_pools)
    filtered_reads = prepare_read_filter(chr_id, saves_prefix, use_filtered_reads)
    gffutils_db = load_genedb(genedb)
    chr_dump_file = saves_file_name(saves_prefix, chr_id)

    return ReadAssignmentLoader(chr_dump_file, gffutils_db, current_chr_record, multimapped_reads, string_pools, filtered_reads)


def create_merging_assignment_loader(chr_id, saves_prefix, string_pools, use_filtered_reads=False):
    multimapped_reads = prepare_multimapped_reads(saves_prefix, chr_id, string_pools)
    filtered_reads = prepare_read_filter(chr_id, saves_prefix, use_filtered_reads)
    chr_dump_file = saves_file_name(saves_prefix, chr_id)

    return MergingSimpleReadAssignmentLoader(chr_dump_file, multimapped_reads, string_pools, filtered_reads)



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
from src.barcode_calling.umi_filtering import (
    UMIFilter,
    format_read_assignment_for_output,

)
from src.isoform_assignment import ReadAssignment, ReadAssignmentType
from src.common import junctions_from_blocks
from src.string_pools import StringPoolManager


class TestFormatReadAssignmentForOutput:
    """Test read assignment output formatting."""
    string_pools = StringPoolManager()

    def test_format_basic(self):
        """Test basic formatting with all attributes."""
        # Create a ReadAssignment
        read_assignment = ReadAssignment(
            read_id="read_001",
            assignment_type=ReadAssignmentType.unique,
            string_pools=self.string_pools
        )
        read_assignment.chr_id = "chr1"
        read_assignment.start = 1000
        read_assignment.end = 2000
        read_assignment.corrected_exons = [(1000, 1500), (1600, 2000)]
        read_assignment.corrected_introns = junctions_from_blocks(read_assignment.corrected_exons)
        read_assignment.barcode = "ACTG"
        read_assignment.umi = "GGGG"
        read_assignment.strand = '+'
        read_assignment.set_additional_attribute('transcript_type', 'known')
        read_assignment.set_additional_attribute('polya_site', 1950)
        read_assignment.set_additional_attribute('cell_type', 'neuron')

        output = format_read_assignment_for_output(read_assignment)

        # Verify all fields are in output
        assert "read_001" in output
        assert "ACTG" in output
        assert "GGGG" in output
        assert "known" in output
        assert "neuron" in output
        # polyA site formatting depends on matching_events, so just check output is non-empty
        assert len(output) > 0

    def test_format_missing_attributes(self):
        """Test formatting with missing optional attributes."""
        read_assignment = ReadAssignment(
            read_id="read_002",
            assignment_type=ReadAssignmentType.ambiguous,
            string_pools=self.string_pools
        )
        read_assignment.chr_id = "chr1"
        read_assignment.start = 1000
        read_assignment.end = 2000
        read_assignment.corrected_exons = [(1000, 2000)]
        read_assignment.corrected_introns = []
        read_assignment.barcode = "ACTG"
        read_assignment.umi = "CCCC"
        read_assignment.strand = '-'

        output = format_read_assignment_for_output(read_assignment)

        # Should handle missing attributes gracefully
        assert "read_002" in output
        assert "unknown" in output or "None" in output


class TestUMIFilter:
    """Test UMIFilter class."""
    string_pools = StringPoolManager()

    @pytest.fixture
    def umi_filter(self):
        """Create a UMIFilter instance."""
        return UMIFilter(umi_length=10, edit_distance=3)

    def test_init(self, umi_filter):
        """Test UMIFilter initialization."""
        assert umi_filter.umi_length == 10
        assert umi_filter.max_edit_distance == 3
        assert isinstance(umi_filter.stats, dict)
        assert isinstance(umi_filter.unique_gene_barcode, set)
        assert isinstance(umi_filter.selected_reads, set)

    def test_construct_umi_dict(self, umi_filter):
        """Test UMI dictionary construction."""
        # Create mock ReadAssignments
        read1 = ReadAssignment(read_id="read_001", assignment_type=ReadAssignmentType.unique, string_pools=self.string_pools)
        read1.umi = "AAAA"

        read2 = ReadAssignment(read_id="read_002", assignment_type=ReadAssignmentType.unique, string_pools=self.string_pools)
        read2.umi = "AAAA"

        read3 = ReadAssignment(read_id="read_003", assignment_type=ReadAssignmentType.unique, string_pools=self.string_pools)
        read3.umi = "TTTT"

        molecule_list = [read1, read2, read3]
        umi_dict = umi_filter._construct_umi_dict(molecule_list)

        # Should group by UMI
        assert len(umi_dict) == 2
        assert "AAAA" in umi_dict
        assert "TTTT" in umi_dict
        assert len(umi_dict["AAAA"]) == 2
        assert len(umi_dict["TTTT"]) == 1

    def test_construct_umi_dict_untrusted(self, umi_filter):
        """Test UMI dict construction with untrusted UMIs."""
        read1 = ReadAssignment(read_id="read_001", assignment_type=ReadAssignmentType.unique, string_pools=self.string_pools)
        read1.umi = ""  # Untrusted UMI

        molecule_list = [read1]
        umi_dict = umi_filter._construct_umi_dict(molecule_list)

        # Untrusted UMIs should be grouped separately
        assert len(umi_dict) == 1
        assert "None" in umi_dict or "" in umi_dict or "read_001" in umi_dict

    def test_select_best_read(self, umi_filter):
        """Test selecting best read from duplicates."""
        # Create reads with different qualities
        read1 = ReadAssignment(read_id="read_001", assignment_type=ReadAssignmentType.unique, string_pools=self.string_pools)
        read1.corrected_exons = [(1000, 1500), (1600, 2000)]

        read2 = ReadAssignment(read_id="read_002", assignment_type=ReadAssignmentType.ambiguous, string_pools=self.string_pools)
        read2.corrected_exons = [(1000, 1500), (1600, 2000)]

        read3 = ReadAssignment(read_id="read_003", assignment_type=ReadAssignmentType.unique, string_pools=self.string_pools)
        read3.corrected_exons = [(1000, 2000)]  # Single exon, less informative

        duplicates = [read2, read1, read3]
        best = umi_filter._process_duplicates(duplicates)

        # Should select unique assignment with more exons
        assert len(best) >= 1
        # The best read should be unique type
        assert any(r.assignment_type == ReadAssignmentType.unique for r in best)


class TestCreateTranscriptInfoDict:
    """Test transcript info dictionary creation."""

    @pytest.mark.skip(reason="Requires actual genedb file, complex to mock")
    def test_create_empty_dict(self):
        """Test creating dict with no genes."""
        # This function requires a real genedb file path, not a dict
        # Skipping as it requires complex setup
        pass

    @pytest.mark.skip(reason="Requires actual genedb file, complex to mock")
    def test_create_dict_with_genes(self):
        """Test creating dict with mock genes."""
        # This function requires a real genedb file path, not a dict
        # Skipping as it requires complex setup
        pass


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Barcode detection classes for single-cell and spatial transcriptomics.

Protocol:
    BarcodeResult: Protocol interface for all detection results

Classes:
    BarcodeDetectionResult: Base result class
    LinkerBarcodeDetectionResult: Result for platforms with linker sequences
    TSOBarcodeDetectionResult: Result for Stereo-seq with TSO detection
    TenXBarcodeDetectionResult: Result for 10x Genomics platform
    SplittingBarcodeDetectionResult: Result for read splitting modes
    ExtractionResult: Dict-based result for universal molecule extraction
    ReadStats: Statistics tracker for barcode detection

Detectors:
    CurioBarcodeDetector: Curio platform detector
    CurioIlluminaDetector: Curio detector for Illumina reads
    StereoBarcodeDetector: Stereo-seq detector
    SharedMemoryStereoBarcodeDetector: Stereo-seq with shared memory
    StereoSplittingBarcodeDetector: Stereo-seq with read splitting
    SharedMemoryStereoSplittingBarcodeDetector: Stereo-seq splitting with shared memory
    SharedMemoryWrapper: Generic shared memory wrapper
    TenXBarcodeDetector: 10x Genomics v3 detector
    TenXv2BarcodeDetector: 10x Genomics v2 detector
    VisiumHDBarcodeDetector: Visium HD detector
    UniversalSingleMoleculeExtractor: Universal barcode detector for custom molecules
"""

from .protocol import BarcodeResult

from .base import (
    BarcodeDetectionResult,
    LinkerBarcodeDetectionResult,
    TSOBarcodeDetectionResult,
    TenXBarcodeDetectionResult,
    SplittingBarcodeDetectionResult,
    ReadStats,
    increase_if_valid,
)

from .extraction_result import (
    ExtractionResult,
    DetectedElement,
)

from .curio import (
    CurioBarcodeDetector,
    CurioIlluminaDetector,
)

from .stereo import (
    StereoBarcodeDetector,
    SharedMemoryStereoBarcodeDetector,
    StereoSplittingBarcodeDetector,
    SharedMemoryStereoSplittingBarcodeDetector,
    SharedMemoryWrapper,
)

from .tenx import (
    TenXBarcodeDetector,
    TenXv2BarcodeDetector,
    VisiumHDBarcodeDetector,
)

from .molecule_structure import MoleculeStructure

from .universal_extraction import UniversalSingleMoleculeExtractor

__all__ = [
    # Protocol
    'BarcodeResult',
    # Result classes
    'BarcodeDetectionResult',
    'LinkerBarcodeDetectionResult',
    'TSOBarcodeDetectionResult',
    'TenXBarcodeDetectionResult',
    'SplittingBarcodeDetectionResult',
    'ExtractionResult',
    'DetectedElement',
    'ReadStats',
    'increase_if_valid',
    # Curio detectors
    'CurioBarcodeDetector',
    'CurioIlluminaDetector',
    # Stereo detectors
    'StereoBarcodeDetector',
    'SharedMemoryStereoBarcodeDetector',
    'StereoSplittingBarcodeDetector',
    'SharedMemoryStereoSplittingBarcodeDetector',
    'SharedMemoryWrapper',
    # 10x detectors
    'TenXBarcodeDetector',
    'TenXv2BarcodeDetector',
    'VisiumHDBarcodeDetector',
    # Universal extraction
    'MoleculeStructure',
    'UniversalSingleMoleculeExtractor',
]




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest

from src.alignment_info import AlignmentInfo
from src.polya_finder import PolyAFinder
from src.polya_verification import PolyAFixer

SEQ1 = 'CTCAAGACCAAGAAGGACGACATGACCATGGCTTAAAAGAGTCTGCTCCCCACAGCCCCCTGCGAT' \
       'GGATGGACGGAGGAACCAGGGTCGGACGACCTCCGATGCTAAGAGCACTCCAACTGCTGCAAACCG' \
       'AAGAAGCAGGCATCGGACDCCTTTTTTTTTTAGGCGCCAAAAAAAAAAAACCAAAAAAAAAAAAA'
TUPLES1 = [(0, 167), (4, 30)]
TUPLES2 = [(0, 167), (4, 30), (5, 5), (4, 16), (0, 30), (5, 70), (0, 201)]
PAIRS = [(0, 75897), (1, 76063), (2, 75995)]


class Alignment:
    def __init__(self, query_name, cigartuples, seq, reference_start, reference_end, reference_name, reference_id,
                 aligned_pairs):
        self.query_name = query_name
        self.cigartuples = cigartuples
        self.seq = seq
        self.reference_start = reference_start
        self.reference_end = reference_end
        self.reference_name = reference_name
        self.reference_id = reference_id
        self.aligned_pairs = aligned_pairs

    def get_aligned_pairs(self):
        return self.aligned_pairs


class TestAlignmentInfo:

    @pytest.mark.parametrize("read_exons, read_blocks, cigar_blocks",
                             [
                                 ([(75897, 76063), (76064, 76294)], [(0, 166), (213, 443)], [(0, 0), (4, 6)])
                             ])
    def test_basic(self, read_exons, read_blocks, cigar_blocks):
        alignment = Alignment('query', TUPLES2, SEQ1, 75896, 113577, 'reference', 789, PAIRS)
        alignment_info = AlignmentInfo(alignment)

        assert alignment_info.alignment == alignment
        assert alignment_info.read_start == read_exons[0][0]
        assert alignment_info.read_end == read_exons[-1][-1]
        assert alignment_info.read_exons == read_exons
        assert alignment_info.read_blocks == read_blocks
        assert alignment_info.cigar_blocks == cigar_blocks

        assert alignment_info.aligned_pairs is None
        assert alignment_info.aligned_pairs_start_index is None
        assert alignment_info.aligned_pairs_end_index is None
        assert alignment_info.polya_info.external_polyt_pos == -1
        assert alignment_info.polya_info.external_polya_pos == -1
        assert alignment_info.polya_info.internal_polya_pos == -1
        assert alignment_info.polya_info.internal_polyt_pos == -1
        assert alignment_info.combined_profile is None
        assert alignment_info.exons_changed is False
        assert len(alignment_info.cage_hits) == 0

    def test_alignment_pairs(self):
        alignment = Alignment('query', TUPLES1, SEQ1, 75896, 113577, 'reference', 789, PAIRS)
        alignment_info = AlignmentInfo(alignment)
        alignment_info.set_aligned_pairs()

        assert len(PAIRS) in alignment_info.aligned_pairs_start_index
        assert len(alignment_info.aligned_pairs_start_index) == 3

    @pytest.mark.parametrize("ref_start, ref_end, intron_index, left_site, chr_record, expected",
                             [
                                 (75999, 81000, None, None, None, (0, 0)),
                                 (75999, 81000, 0, False, None, (0, 0)),
                                 (75999, 81000, 0, True, None, (0, 0))
                             ])
    def test_get_error_count(self, ref_start, ref_end, intron_index, left_site, chr_record, expected):
        alignment = Alignment('query', TUPLES2, SEQ1, 75896, 113577, 'reference', 789, PAIRS)
        alignment_info = AlignmentInfo(alignment)

        indel_count, mismatch_count = alignment_info.get_error_count(ref_start, ref_end, intron_index, left_site,
                                                                     chr_record)
        assert (indel_count, mismatch_count) == expected

    @pytest.mark.parametrize("read_start, read_end",
                             [
                                 (51055, 51221)
                             ])
    def test_add_polya_info(self, read_start, read_end):
        alignment = Alignment('query', TUPLES2, SEQ1, 51054, 35000, 'reference', 789, PAIRS)
        alignment_info = AlignmentInfo(alignment)
        polya_finder = PolyAFinder()
        polya_fixer = PolyAFixer(1)

        alignment_info.add_polya_info(polya_finder, polya_fixer)

        assert alignment_info.exons_changed is True
        assert alignment_info.polya_info.internal_polya_pos == read_end
        assert alignment_info.polya_info.internal_polyt_pos == -1
        assert alignment_info.polya_info.external_polya_pos == -1
        assert alignment_info.polya_info.external_polyt_pos == -1
        assert alignment_info.read_exons == [(read_start, read_end)]
        assert alignment_info.read_blocks == [(0, (read_end - read_start))]
        assert alignment_info.cigar_blocks == [(0, 0)]
        assert alignment_info.read_start == read_start
        assert alignment_info.read_end == read_end

    def test_get_seq(self):
        alignment = Alignment('query', TUPLES1, SEQ1, 75896, 113577, 'reference', 789, PAIRS)
        alignment_info = AlignmentInfo(alignment)

        with pytest.raises(AssertionError):
            alignment_info.get_seq(76000, 81000)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import pybedtools as pbt


logger = logging.getLogger('IsoQuant')


class CagePeakFinder:
    def __init__(self, cage_file, shift_size=50, window_size=5):
        self.cage_peaks = self._load_cage_peaks(cage_file)
        self.shift_size = shift_size
        self.window_size = window_size

    def _load_cage_peaks(self, cage_file):
        return pbt.BedTool(cage_file)

    def _get_search_region(self, alignment, extended=False):
        contig = alignment.reference_name
        search_size = self.shift_size if extended else self.window_size
        if alignment.is_reverse:
            strand = '-'
            start = max(alignment.reference_end - self.window_size, 0)
            end = alignment.reference_end + search_size
        else:
            strand = '.'
            start = max(alignment.reference_start - search_size, 0)
            end = alignment.reference_start + self.window_size
        return contig, start, end, strand

    def find_cage_peak(self, alignment):
        logger.debug("Searching for cage peak for %s " % alignment.query_name)

        contig, start, end, strand = self._get_search_region(alignment, extended=True)
        alignment_interval = pbt.Interval(chrom=contig, start=start, end=end, strand=strand)
        cage_intersections = self.cage_peaks.all_hits(alignment_interval)

        if len(cage_intersections) > 0:
            logger.debug('CAGE peaks found: {}'.format(cage_intersections))
        else:
            logger.debug('No CAGE peaks found')

        return cage_intersections



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import unittest
from src.file_naming import *


class TestFileNaming(unittest.TestCase):
    def test_saves_file_name(self):
        base_name = "test_output"
        chr_id = "chr1"
        result = saves_file_name(base_name, chr_id)
        self.assertIn(base_name, result)

    def test_multimappers_file_name(self):
        base_name = "test_output"
        chr_id = "chr1"
        result = multimappers_file_name(base_name, chr_id)
        self.assertIn(base_name, result)
        self.assertIn("multimappers", result)
        
    def test_filtered_reads_file_name(self):
        base_name = "test_output"
        chr_id = "chr1"
        result = filtered_reads_file_name(base_name, chr_id)
        self.assertIn(base_name, result)
        self.assertIn("filtered", result)
        
    def test_chromosome_id_sanitization(self):
        # Test that special characters in chromosome IDs are handled
        chr_id = "chr1:1000-2000"
        result = saves_file_name("test", chr_id)
        # Should not contain colons or dashes in unexpected places
        self.assertIsInstance(result, str)
        self.assertTrue(len(result) > 0)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import unittest
import threading
from unittest.mock import MagicMock, patch
from src.id_policy import (
    SimpleIDDistributor,
    ExcludingIdDistributor,
    FeatureIdStorage,
    AtomicIDDistributor
)


class TestSimpleIDDistributor(unittest.TestCase):
    def test_get_next_id(self):
        distributor = SimpleIDDistributor()
        self.assertEqual(distributor.increment(), 1)
        self.assertEqual(distributor.increment(), 2)
        self.assertEqual(distributor.increment(), 3)

    def test_initial_value(self):
        distributor = SimpleIDDistributor()
        self.assertEqual(distributor.value, 0)

    def test_sequential_increment(self):
        distributor = SimpleIDDistributor()
        # Generate 100 IDs
        ids = [distributor.increment() for _ in range(100)]
        # Should be sequential 1 to 100
        self.assertEqual(ids, list(range(1, 101)))


class TestExcludingIdDistributor(unittest.TestCase):
    def test_without_genedb(self):
        """Test ExcludingIdDistributor works without genedb."""
        distributor = ExcludingIdDistributor(None, "chr1")
        self.assertEqual(distributor.increment(), 1)
        self.assertEqual(distributor.increment(), 2)
        self.assertEqual(distributor.increment(), 3)

    def test_skips_forbidden_gene_ids(self):
        """Test that forbidden gene IDs from genedb are skipped."""
        # Create mock genedb
        mock_genedb = MagicMock()

        # Mock gene with novel gene prefix (novel_gene_5 per TranscriptNaming)
        mock_gene = MagicMock()
        mock_gene.id = "novel_gene_5"

        # Mock transcript with novel transcript prefix (transcript10.1 per TranscriptNaming)
        mock_transcript = MagicMock()
        mock_transcript.id = "transcript10.1"

        # region() returns genes and transcripts
        # Note: for transcripts, the code passes a tuple ("transcript", "mRNA")
        def mock_region(seqid, start, featuretype):
            if featuretype == "gene":
                return [mock_gene]
            elif featuretype == ("transcript", "mRNA"):
                return [mock_transcript]
            return []

        mock_genedb.region = mock_region

        distributor = ExcludingIdDistributor(mock_genedb, "chr1")

        # Should skip 5 and 10
        self.assertIn(5, distributor.forbidden_ids)
        self.assertIn(10, distributor.forbidden_ids)

        # Generate IDs - should skip 5 and 10
        ids = []
        for _ in range(12):
            ids.append(distributor.increment())

        self.assertNotIn(5, ids)
        self.assertNotIn(10, ids)
        # First few IDs should be 1, 2, 3, 4, 6, 7, 8, 9, 11, 12, 13, 14
        self.assertEqual(ids[:4], [1, 2, 3, 4])
        self.assertEqual(ids[4], 6)  # Skipped 5
        self.assertEqual(ids[8], 11)  # Skipped 10

    def test_handles_malformed_gene_ids(self):
        """Test that malformed gene IDs don't crash."""
        mock_genedb = MagicMock()

        mock_gene1 = MagicMock()
        mock_gene1.id = "ISOQUANT_gene_abc"  # Not a number

        mock_gene2 = MagicMock()
        mock_gene2.id = "ISOQUANT_gene_"  # Missing number

        mock_gene3 = MagicMock()
        mock_gene3.id = "other_gene_123"  # Wrong prefix

        def mock_region(seqid, start, featuretype):
            if featuretype == "gene":
                return [mock_gene1, mock_gene2, mock_gene3]
            return []

        mock_genedb.region = mock_region

        # Should not raise exception
        distributor = ExcludingIdDistributor(mock_genedb, "chr1")

        # Forbidden IDs should be empty (all malformed)
        self.assertEqual(len(distributor.forbidden_ids), 0)


class TestFeatureIdStorage(unittest.TestCase):
    def test_without_genedb(self):
        """Test FeatureIdStorage works without genedb."""
        id_distributor = SimpleIDDistributor()
        storage = FeatureIdStorage(id_distributor)

        # First call for new feature returns int (from increment())
        id1 = storage.get_id("chr1", (100, 200), "+")
        self.assertEqual(id1, 1)  # Returns int from increment()

        id2 = storage.get_id("chr1", (300, 400), "+")
        self.assertEqual(id2, 2)  # Returns int from increment()

        # Second call for same feature returns string (from id_dict)
        id3 = storage.get_id("chr1", (100, 200), "+")
        self.assertEqual(id3, "chr1.1")  # Returns stored string

    def test_same_feature_cached_id(self):
        """Test that same feature coordinates return cached ID on second call."""
        id_distributor = SimpleIDDistributor()
        storage = FeatureIdStorage(id_distributor)

        # First call returns int
        id1 = storage.get_id("chr1", (100, 200), "+")
        self.assertEqual(id1, 1)

        # Second call returns cached string
        id2 = storage.get_id("chr1", (100, 200), "+")
        self.assertEqual(id2, "chr1.1")

    def test_different_strand_different_id(self):
        """Test that same coordinates but different strand get different IDs."""
        id_distributor = SimpleIDDistributor()
        storage = FeatureIdStorage(id_distributor)

        id1 = storage.get_id("chr1", (100, 200), "+")
        id2 = storage.get_id("chr1", (100, 200), "-")

        # Both are first calls, so both return int
        self.assertEqual(id1, 1)
        self.assertEqual(id2, 2)
        self.assertNotEqual(id1, id2)

    def test_different_chr_different_id(self):
        """Test that same coordinates on different chromosomes get different IDs."""
        id_distributor = SimpleIDDistributor()
        storage = FeatureIdStorage(id_distributor)

        id1 = storage.get_id("chr1", (100, 200), "+")
        id2 = storage.get_id("chr2", (100, 200), "+")

        # Both are first calls, so both return int
        self.assertEqual(id1, 1)
        self.assertEqual(id2, 2)
        self.assertNotEqual(id1, id2)

    def test_with_genedb_preloaded_exons(self):
        """Test that existing exon IDs from genedb are preserved."""
        mock_genedb = MagicMock()

        # Mock existing exon feature
        mock_exon = MagicMock()
        mock_exon.start = 100
        mock_exon.end = 200
        mock_exon.strand = "+"
        mock_exon.attributes = {"exon_id": ["existing_exon_1"]}

        def mock_region(seqid, start, featuretype):
            if featuretype == "exon":
                return [mock_exon]
            return []

        mock_genedb.region = mock_region

        id_distributor = SimpleIDDistributor()
        storage = FeatureIdStorage(id_distributor, mock_genedb, "chr1", "exon")

        # Same coordinates should return preloaded ID (already in id_dict)
        existing_id = storage.get_id("chr1", (100, 200), "+")
        self.assertEqual(existing_id, "existing_exon_1")

        # New coordinates should generate new ID (first call returns int)
        new_id = storage.get_id("chr1", (300, 400), "+")
        self.assertEqual(new_id, 1)  # Returns int from increment()


class TestAtomicIDDistributor(unittest.TestCase):
    def test_basic_increment(self):
        """Test basic increment functionality."""
        distributor = AtomicIDDistributor()
        self.assertEqual(distributor.increment(), 1)
        self.assertEqual(distributor.increment(), 2)
        self.assertEqual(distributor.increment(), 3)

    def test_thread_safety(self):
        """Test that AtomicIDDistributor is thread-safe."""
        distributor = AtomicIDDistributor()
        results = []
        num_threads = 10
        increments_per_thread = 100

        def worker():
            for _ in range(increments_per_thread):
                results.append(distributor.increment())

        threads = [threading.Thread(target=worker) for _ in range(num_threads)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # All IDs should be unique
        self.assertEqual(len(results), num_threads * increments_per_thread)
        self.assertEqual(len(set(results)), len(results))

        # Should have generated IDs 1 through 1000
        self.assertEqual(sorted(results), list(range(1, num_threads * increments_per_thread + 1)))


if __name__ == '__main__':
    unittest.main()




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import os
import subprocess
from collections import namedtuple

import pytest

from src.common import get_path_to_program


class TestGetPathToProgram:
    def file_exist(self, path):
        return True

    def file_doesnot_exist(self, path):
        return False

    def is_access_true(self, path, mode):
        return True

    def test_in_dir(self, monkeypatch):
        def correct_versions(*args, **kwargs):
            out = namedtuple("CompletedProcess", ("stdout"))
            return out("1.38")

        monkeypatch.setattr(subprocess, "run", correct_versions)
        monkeypatch.setattr(os.path, "isfile", self.file_exist)
        monkeypatch.setattr(os, "access", self.is_access_true)
        program_name = "my_prog"
        dir = "/folder"
        assert os.path.join(dir, program_name) == get_path_to_program(program_name, dir, "1.38")

    @pytest.mark.parametrize("min_version", ["1.39", "2.0"])
    def test_wrong_version(self, monkeypatch, min_version):
        def correct_versions(*args, **kwargs):
            out = namedtuple("CompletedProcess", ("stdout"))
            return out("1.38")

        monkeypatch.setattr(subprocess, "run", correct_versions)
        monkeypatch.setattr(os.path, "isfile", self.file_exist)
        monkeypatch.setattr(os, "access", self.is_access_true)
        program_name = "my_prog"
        dir = "/folder"
        assert None is get_path_to_program(program_name, dir, "1.39")

    @pytest.mark.parametrize("min_version", ["1.39", "2.0"])
    def test_not_in_path(self, monkeypatch, min_version):
        monkeypatch.setenv("PATH", os.pathsep.join(["path1", "path2", "path3"]))
        monkeypatch.setattr(os.path, "isfile", self.file_doesnot_exist)
        monkeypatch.setattr(os, "access", self.is_access_true)
        program_name = "my_prog"
        assert None is get_path_to_program(program_name, "1.39")

    # TODO: add more tests



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
String interning for memory optimization.

Replaces duplicated strings with integer indices referencing shared string pools.
Expected memory reduction: ~78% for billion-read datasets.

Architecture:
- StringPool: Bidirectional mapping between strings and integers
- StringPoolManager: Manages all pools for a worker
  - Global pools: gene, transcript (from annotation), chromosome (from chr_ids list)
  - Per-chromosome pools: barcode, UMI (from split files)
  - Read group pools: Multiple types depending on strategy

Usage:
- Workers build chromosome pool from chr_ids list
- Workers build gene/transcript pools from annotation database
- All IsoformMatch/ReadAssignment objects store integer IDs
- Properties provide transparent string access via pools
- Serialization writes integer format for disk savings

Implementation Details:
- All data classes (IsoformMatch, ReadAssignment, BasicReadAssignment)
  require string_pools parameter - no fallback to strings
- Pools are built deterministically (sorted order) for consistency
- Per-worker pools ensure no sharing/locking between parallel workers
"""

import os
import glob
import re
import logging
from typing import List, Dict

from .read_groups import AbstractReadGrouper, get_grouping_pool_types
from .assignment_loader import load_genedb
from .serialization import write_int, write_string, read_int, read_string

logger = logging.getLogger('IsoQuant')


class StringPool:
    """
    Bidirectional mapping between strings and integers for memory optimization.

    Each unique string is stored once and referenced by a unique integer ID.
    Provides O(1) lookup in both directions.
    """

    def __init__(self):
        """Initialize empty string pool."""
        self.str_to_int: Dict[str, int] = {}
        self.int_to_str: List[str] = []
        self.next_id: int = 0

    def add(self, string: str) -> int:
        """
        Add string to pool and return its integer ID.
        If string already exists, returns existing ID.

        Args:
            string: String to add to pool

        Returns:
            Integer ID for the string
        """
        if string in self.str_to_int:
            return self.str_to_int[string]

        int_id = self.next_id
        self.str_to_int[string] = int_id
        self.int_to_str.append(string)
        self.next_id += 1
        return int_id

    def get_int(self, string: str) -> int:
        """
        Get integer ID for string. Adds string if not present.

        Args:
            string: String to look up

        Returns:
            Integer ID for the string
        """
        return self.add(string)

    def get_str(self, int_id: int) -> str:
        """
        Get string for integer ID.

        Args:
            int_id: Integer ID to look up

        Returns:
            String corresponding to the ID

        Raises:
            IndexError: If int_id is out of range
        """
        return self.int_to_str[int_id]

    def __len__(self) -> int:
        """Return number of unique strings in pool."""
        return len(self.int_to_str)

    def __contains__(self, string: str) -> bool:
        """Check if string is in pool."""
        return string in self.str_to_int


class StringPoolManager:
    """
    Manages all string pools for a chromosome processing worker.

    Pools are categorized by data type:
    - Global pools: gene, transcript (from annotation), chromosome (from chr_ids list)
    - Per-chromosome pools: barcode, UMI (from split files)
    - Read group pools: Multiple types depending on grouping strategy
    """

    def __init__(self):
        """Initialize all string pools."""
        # Global pools
        self.gene_pool = StringPool()  # from annotation
        self.transcript_pool = StringPool()  # from annotation
        self.chromosome_pool = StringPool()  # from chr_ids list

        # Per-chromosome pools (from split files)
        self.barcode_pool = StringPool()
        self.umi_pool = StringPool()

        # Read group pools by type
        self.read_group_tsv_pools: Dict[str, StringPool] = {}  # pool_key -> pool (static, from TSV files)
        self.read_group_dynamic_pools: Dict[int, StringPool] = {}  # spec_index -> pool (dynamic, for tag/read_id)
        self.file_name_pool = StringPool()  # For file_name grouping (static)
        self.barcode_spot_pools: Dict[int, StringPool] = {}  # For barcode_spot grouping (multi-column support)

        # Mapping from group spec index to pool type ('file_name', 'barcode_spot', 'tsv:N', 'dynamic')
        self.group_spec_pool_types: Dict[int, str] = {}

    def build_from_gffutils(self, gffutils_db):
        """
        Build global pools from gffutils database.

        Pools are built in deterministic sorted order to ensure consistency
        across workers.

        Args:
            gffutils_db: gffutils.FeatureDB object
        """
        logger.debug("Building string pools from gffutils database")

        # Gene pool
        gene_ids = set()
        for gene in gffutils_db.features_of_type('gene'):
            gene_ids.add(gene.id)
        for gene_id in sorted(gene_ids):
            self.gene_pool.add(gene_id)
        logger.debug(f"Gene pool: {len(self.gene_pool)} unique genes")

        # Transcript pool
        transcript_ids = set()
        # Query different transcript feature types
        for feature_type in ['transcript', 'RNA', 'mRNA', 'miRNA', 'ncRNA', 'tRNA', 'rRNA',
                             'snRNA', 'lnc_RNA', 'snoRNA', 'telomerase_RNA', 'Y_RNA', 'scRNA']:
            try:
                for transcript in gffutils_db.features_of_type(feature_type):
                    transcript_ids.add(transcript.id)
            except:
                pass  # Feature type might not exist
        for transcript_id in sorted(transcript_ids):
            self.transcript_pool.add(transcript_id)
        logger.debug(f"Transcript pool: {len(self.transcript_pool)} unique transcripts")

    def build_from_gene_db(self, gene_db):
        """
        Build global pools from gene annotation database.

        Pools are built in deterministic sorted order to ensure consistency
        across workers.

        Args:
            gene_db: GeneInfo database from gene_info.py
        """
        logger.debug("Building string pools from gene database")

        # Gene pool
        gene_ids = sorted(gene_db.genes.keys())
        for gene_id in gene_ids:
            self.gene_pool.add(gene_id)
        logger.debug(f"Gene pool: {len(self.gene_pool)} unique genes")

        # Transcript pool
        transcript_ids = sorted(gene_db.transcripts.keys())
        for transcript_id in transcript_ids:
            self.transcript_pool.add(transcript_id)
        logger.debug(f"Transcript pool: {len(self.transcript_pool)} unique transcripts")

    def build_chromosome_pool(self, chr_ids):
        """
        Build chromosome pool from list of chromosome IDs.

        Args:
            chr_ids: List of chromosome IDs to process
        """
        # Sort alphabetically for deterministic order across workers
        for chr_id in sorted(chr_ids):
            self.chromosome_pool.add(chr_id)
        logger.debug(f"Chromosome pool: {len(self.chromosome_pool)} chromosomes")

    def build_file_name_pool(self, sample):
        """
        Build file name pool from sample file list.

        Uses same logic as FileNameGrouper: uses readable_names_dict (labels)
        if available, otherwise uses basename without extension.
        Pool is built in sorted order for deterministic IDs across workers.
        Also includes 'NA' for reads without filename info.

        Args:
            sample: Sample object with file_list and readable_names_dict attributes
        """
        # Collect all file names first
        file_names = set()

        # If readable_names_dict is available (e.g., when -l labels are used),
        # use its values - matching FileNameGrouper behavior
        if sample.readable_names_dict:
            for readable_name in sample.readable_names_dict.values():
                file_names.add(readable_name)
        else:
            # file_list is a list of lists (libraries), each library has one or more files
            for lib in sample.file_list:
                # Get basename without extension, matching FileNameGrouper logic
                readable_name = os.path.splitext(os.path.basename(lib[0]))[0]
                file_names.add(readable_name)

        # Add in sorted order for deterministic IDs
        for name in sorted(file_names):
            self.file_name_pool.add(name)
        logger.debug(f"File name pool: {len(self.file_name_pool)} unique readable names")

    def build_barcode_spot_pool(self, barcode2spot_spec: str, column_index: int = 0,
                                 file_column: int = 1):
        """
        Build barcode-to-spot pool for a specific column.

        Pool is built in sorted order for deterministic IDs across workers.
        Also includes 'NA' for reads without barcode mapping.

        Args:
            barcode2spot_spec: barcode2spot spec (may include :col:cols suffix)
            column_index: 0-based pool index (for storage in barcode_spot_pools dict)
            file_column: Actual file column to read (from parsed spec)
        """
        # Extract filename from spec
        bc2spot_file = barcode2spot_spec.split(':')[0]

        # Collect all spots for this column
        spots = set()
        with open(bc2spot_file, 'r') as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                parts = line.split('\t')
                if len(parts) > file_column:
                    spots.add(parts[file_column])

        # Add default group ID for reads without barcode mapping
        spots.add(AbstractReadGrouper.default_group_id)

        # Create and populate pool
        pool = StringPool()
        for spot in sorted(spots):
            pool.add(spot)

        self.barcode_spot_pools[column_index] = pool
        logger.debug(f"Barcode spot pool (col {column_index}): {len(pool)} unique spots")

    def load_barcode_pool(self, barcode_file: str):
        """
        Load barcode pool from split barcode file.

        Pool is built in sorted order for deterministic IDs across workers.
        Also includes 'NA' for reads without barcode.

        Args:
            barcode_file: Path to split barcode file for chromosome
                         Format: read_id\tbarcode\tumi
        """
        if not os.path.exists(barcode_file):
            logger.debug(f"Barcode file not found: {barcode_file}")
            return

        # Collect all barcodes and UMIs first
        # Use split('\t') to preserve empty fields (empty UMI = consecutive tabs)
        barcodes = set()
        umis = set()
        with open(barcode_file, 'r') as f:
            for line in f:
                if line.startswith('#') or line.startswith('read_id\t'):
                    continue
                parts = line.strip().split('\t')
                if len(parts) >= 2 and parts[1]:
                    barcodes.add(parts[1])
                if len(parts) >= 3 and parts[2]:
                    umis.add(parts[2])

        # Add default group ID for reads without barcode
        barcodes.add(AbstractReadGrouper.default_group_id)

        # Add in sorted order for deterministic IDs
        for barcode in sorted(barcodes):
            self.barcode_pool.add(barcode)
        for umi in sorted(umis):
            self.umi_pool.add(umi)

        logger.debug(f"Barcode pool: {len(self.barcode_pool)} unique barcodes")
        logger.debug(f"UMI pool: {len(self.umi_pool)} unique UMIs")

    def load_read_group_tsv_pool(self, read_group_file: str, pool_key: str, col_index: int = 1, delimiter: str = '\t'):
        """
        Load read group pool from split TSV/CSV file.

        Pool is built in sorted order for deterministic IDs across workers.

        Args:
            read_group_file: Path to split read_group file for chromosome
            pool_key: Unique key for this pool (e.g., 'spec2_col1' for multi-column files)
            col_index: Column index to read group values from (1-based, default 1)
            delimiter: Field delimiter (from file spec, defaults to tab)
        """
        if not os.path.exists(read_group_file):
            logger.debug(f"Read group file not found: {read_group_file}")
            return

        # Collect all group values first
        # Always include 'NA' (default group ID) in case some reads are not in the TSV file
        group_values = {AbstractReadGrouper.default_group_id}
        with open(read_group_file, 'r') as f:
            for line in f:
                if line.startswith('#'):
                    continue
                line = line.strip()
                if not line:
                    continue
                parts = line.split(delimiter)
                if len(parts) > col_index:
                    group_values.add(parts[col_index])

        # Add in sorted order for deterministic IDs
        pool = StringPool()
        for value in sorted(group_values):
            pool.add(value)

        self.read_group_tsv_pools[pool_key] = pool
        logger.debug(f"Read group TSV pool ({pool_key}): {len(pool)} unique values")

    def set_group_spec_pool_type(self, spec_index: int, pool_type: str):
        """
        Set the pool type for a grouping specification index.

        Args:
            spec_index: Index of the grouping spec (position in read_group list)
            pool_type: One of 'file_name', 'barcode_spot', 'tsv:N', 'dynamic'
        """
        self.group_spec_pool_types[spec_index] = pool_type

    def get_read_group_pool(self, spec_index: int) -> StringPool:
        """
        Get the appropriate pool for a grouping spec index.

        Args:
            spec_index: Index of the grouping spec

        Returns:
            The StringPool to use for this grouping spec
        """
        if spec_index not in self.group_spec_pool_types:
            # Default to dynamic pool if not specified
            return self._get_or_create_dynamic_pool(spec_index)

        pool_type = self.group_spec_pool_types[spec_index]

        if pool_type == 'file_name':
            return self.file_name_pool
        elif pool_type.startswith('barcode_spot:'):
            # Multi-column barcode_spot: 'barcode_spot:COL_INDEX'
            col_idx = int(pool_type.split(':')[1])
            if col_idx in self.barcode_spot_pools:
                return self.barcode_spot_pools[col_idx]
            return self._get_or_create_dynamic_pool(spec_index)
        elif pool_type == 'barcode_spot':
            # Legacy single-column (use column 0, first spot column)
            if 0 in self.barcode_spot_pools:
                return self.barcode_spot_pools[0]
            return self._get_or_create_dynamic_pool(spec_index)
        elif pool_type == 'barcode':
            # Use barcode pool for direct barcode grouping
            return self.barcode_pool
        elif pool_type.startswith('tsv:'):
            # Extract pool_key from 'tsv:SPEC_INDEX:COL_INDEX:DELIMITER'
            # pool_key format is 'SPEC_INDEX:COL_INDEX'
            parts = pool_type.split(':')
            if len(parts) >= 3:
                pool_key = f"{parts[1]}:{parts[2]}"
            else:
                pool_key = parts[1]  # Fallback for old format 'tsv:N'
            if pool_key in self.read_group_tsv_pools:
                return self.read_group_tsv_pools[pool_key]
            else:
                # Pool not loaded yet, use dynamic pool for this spec
                return self._get_or_create_dynamic_pool(spec_index)
        else:  # 'dynamic' or unknown
            return self._get_or_create_dynamic_pool(spec_index)

    def _get_or_create_dynamic_pool(self, spec_index: int) -> StringPool:
        """Get or create a dynamic pool for a spec index."""
        if spec_index not in self.read_group_dynamic_pools:
            self.read_group_dynamic_pools[spec_index] = StringPool()
        return self.read_group_dynamic_pools[spec_index]

    def read_group_to_ids(self, group_strings: List[str]) -> List[int]:
        """
        Convert list of group strings to list of integer IDs.

        Args:
            group_strings: List of group ID strings

        Returns:
            List of integer IDs
        """
        if not group_strings:
            return []

        ids = []
        for spec_index, group_str in enumerate(group_strings):
            if group_str is not None:
                pool = self.get_read_group_pool(spec_index)
                ids.append(pool.get_int(group_str))
            else:
                ids.append(-1)  # Use -1 for None values
        return ids

    def read_group_from_ids(self, group_ids: List[int]) -> List[str]:
        """
        Convert list of integer IDs to list of group strings.

        Args:
            group_ids: List of integer IDs

        Returns:
            List of group ID strings
        """
        if not group_ids:
            return []

        strings = []
        for spec_index, group_id in enumerate(group_ids):
            if group_id < 0:
                strings.append(None)  # -1 means None
            else:
                pool = self.get_read_group_pool(spec_index)
                try:
                    strings.append(pool.get_str(group_id))
                except IndexError:
                    # Pool not populated yet or ID out of range
                    # This can happen during deserialization if pools weren't loaded
                    # (e.g., default grouper with 'NA' or dynamic pools not yet built)
                    # Fall back to returning the ID as a string
                    pool_type = self.group_spec_pool_types.get(spec_index, 'unknown')
                    logger.warning(f"Read group ID {group_id} not found in pool for spec {spec_index} (type={pool_type}, pool_size={len(pool)}), using ID as string")
                    strings.append(str(group_id))
        return strings

    def has_dynamic_pools(self) -> bool:
        """Check if any dynamic pools have data."""
        return any(len(pool) > 0 for pool in self.read_group_dynamic_pools.values())

    def serialize_dynamic_pools(self, outfile):
        """
        Serialize dynamic read group pools to binary file.

        Format:
            num_specs (int)
            for each spec:
                spec_index (int)
                num_strings (int)
                for each string:
                    string (length-prefixed)
        """

        # Only serialize non-empty dynamic pools
        non_empty_pools = {idx: pool for idx, pool in self.read_group_dynamic_pools.items()
                          if len(pool) > 0}

        write_int(len(non_empty_pools), outfile)
        for spec_index in sorted(non_empty_pools.keys()):
            pool = non_empty_pools[spec_index]
            write_int(spec_index, outfile)
            write_int(len(pool), outfile)
            # Write strings in ID order (0, 1, 2, ...) to preserve mapping
            for i in range(len(pool)):
                write_string(pool.get_str(i), outfile)

    def deserialize_dynamic_pools(self, infile):
        """
        Deserialize dynamic read group pools from binary file.

        Rebuilds pools with same ID mappings as when serialized.
        """
        num_specs = read_int(infile)
        for _ in range(num_specs):
            spec_index = read_int(infile)
            num_strings = read_int(infile)
            pool = StringPool()
            for _ in range(num_strings):
                string = read_string(infile)
                pool.add(string)  # IDs assigned in order: 0, 1, 2, ...
            self.read_group_dynamic_pools[spec_index] = pool

    def save_barcode_umi_pools(self, filepath: str):
        """Save barcode and UMI pools to binary file, preserving ID order."""
        with open(filepath, 'wb') as f:
            write_int(len(self.barcode_pool), f)
            for i in range(len(self.barcode_pool)):
                write_string(self.barcode_pool.get_str(i), f)
            write_int(len(self.umi_pool), f)
            for i in range(len(self.umi_pool)):
                write_string(self.umi_pool.get_str(i), f)

    def load_barcode_umi_pools(self, filepath: str):
        """Load barcode and UMI pools from binary file, preserving ID order."""
        if not os.path.exists(filepath):
            return
        with open(filepath, 'rb') as f:
            num_barcodes = read_int(f)
            for _ in range(num_barcodes):
                self.barcode_pool.add(read_string(f))
            num_umis = read_int(f)
            for _ in range(num_umis):
                self.umi_pool.add(read_string(f))

    def get_stats(self) -> str:
        """
        Get statistics about all pools.

        Returns:
            Formatted string with pool sizes
        """
        lines = ["String Pool Statistics:"]
        lines.append(f"  Genes: {len(self.gene_pool)}")
        lines.append(f"  Transcripts: {len(self.transcript_pool)}")
        lines.append(f"  Chromosomes: {len(self.chromosome_pool)}")
        lines.append(f"  Barcodes: {len(self.barcode_pool)}")
        lines.append(f"  UMIs: {len(self.umi_pool)}")
        lines.append(f"  File names: {len(self.file_name_pool)}")
        total_spots = sum(len(p) for p in self.barcode_spot_pools.values())
        lines.append(f"  Barcode spots: {total_spots} (across {len(self.barcode_spot_pools)} pools)")
        for spec_idx, pool in self.read_group_dynamic_pools.items():
            lines.append(f"  Read group dynamic (spec {spec_idx}): {len(pool)}")
        for spec_idx, pool in self.read_group_tsv_pools.items():
            lines.append(f"  Read group TSV (spec {spec_idx}): {len(pool)}")
        return "\n".join(lines)


def setup_string_pools(args, sample, chr_ids, chr_id=None, gffutils_db=None,
                       load_barcode_pool=False, load_tsv_pools=False, read_group_file_prefix=None):
    """
    Set up string pools for memory optimization during parallel processing.

    Args:
        args: Command-line arguments
        sample: SampleData object
        chr_ids: List of all chromosome IDs to process (for chromosome pool)
        chr_id: Current chromosome ID (required if loading per-chromosome pools)
        gffutils_db: Pre-loaded gffutils database (if None, will load from args.genedb)
        load_barcode_pool: Whether to load per-chromosome barcode pool
        load_tsv_pools: Whether to load per-chromosome TSV read group pools
        read_group_file_prefix: Override for read_group_file path (used when sample has modified prefix)

    Returns:
        StringPoolManager with pools configured
    """

    string_pools = StringPoolManager()

    # Build chromosome pool from the definitive list of chromosomes to process
    string_pools.build_chromosome_pool(chr_ids)

    # Build gene/transcript pools from annotation (if available)
    if gffutils_db is None and args.genedb:
        gffutils_db = load_genedb(args.genedb)
    if gffutils_db:
        string_pools.build_from_gffutils(gffutils_db)

    # Set up read group pool type mapping
    pool_types = get_grouping_pool_types(args)
    for spec_idx, pool_type in pool_types.items():
        string_pools.set_group_spec_pool_type(spec_idx, pool_type)

    # Build file_name pool if needed
    if any(pt == 'file_name' for pt in pool_types.values()):
        string_pools.build_file_name_pool(sample)

    # Build barcode_spot pools if needed (supports multi-column)
    if any(pt == 'barcode_spot' or pt.startswith('barcode_spot:') for pt in pool_types.values()):
        if args.barcode2spot:
            from .read_groups import parse_barcode2spot_spec
            filename, barcode_col, spot_cols = parse_barcode2spot_spec(args.barcode2spot)
            for col_idx, file_col in enumerate(spot_cols):
                string_pools.build_barcode_spot_pool(args.barcode2spot, column_index=col_idx,
                                                      file_column=file_col)

    # Load per-chromosome pools if requested
    if chr_id:
        if load_barcode_pool and sample.barcodes_split_reads:
            barcode_file = sample.get_barcodes_split_file(chr_id)
            string_pools.load_barcode_pool(barcode_file)
            # For --barcoded_bam mode: load from binary pool file saved during collect phase
            if len(string_pools.barcode_pool) == 0 and getattr(args, 'barcoded_bam', False):
                string_pools.load_barcode_umi_pools(sample.get_barcode_pools_file(chr_id))

        # Use override if provided, otherwise use sample's read_group_file
        read_group_file = read_group_file_prefix if read_group_file_prefix else sample.read_group_file
        if load_tsv_pools and read_group_file:
            # Build mapping from tsv spec_index to (col_index, delimiter) from pool_types
            # pool_type format: 'tsv:SPEC_INDEX:COL_INDEX:DELIMITER'
            tsv_pool_info = {}  # spec_index -> list of (col_index, delimiter, pool_key)
            for pool_type in pool_types.values():
                if pool_type.startswith('tsv:'):
                    parts = pool_type.split(':')
                    if len(parts) >= 4:
                        tsv_spec_idx = int(parts[1])
                        col_idx = int(parts[2])
                        delimiter = parts[3]
                        # pool_key is spec_idx:col_idx to uniquely identify multi-column pools
                        pool_key = f"{tsv_spec_idx}:{col_idx}"
                        if tsv_spec_idx not in tsv_pool_info:
                            tsv_pool_info[tsv_spec_idx] = []
                        tsv_pool_info[tsv_spec_idx].append((col_idx, delimiter, pool_key))

            base_pattern = read_group_file + "_spec*_" + chr_id
            spec_files = sorted(glob.glob(base_pattern))
            for spec_file in spec_files:
                match = re.search(r'_spec(\d+)_', spec_file)
                if match:
                    spec_index = int(match.group(1))
                    # Load pool for each column in this spec file
                    for col_idx, delimiter, pool_key in tsv_pool_info.get(spec_index, [(1, '\t', f"{spec_index}:1")]):
                        string_pools.load_read_group_tsv_pool(spec_file, pool_key, col_idx, delimiter)

    return string_pools



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Base classes for barcode detection results.

Provides result containers for barcode calling across different platforms.
"""

from collections import defaultdict
from typing import List, Optional, Dict, Iterable


def increase_if_valid(val: Optional[int], delta: int) -> Optional[int]:
    """
    Increment a coordinate value if it's valid.

    Args:
        val: Position value (-1 or None indicates invalid)
        delta: Amount to increment

    Returns:
        Incremented value if valid, otherwise original value
    """
    if val and val != -1:
        return val + delta
    return val





class BarcodeDetectionResult:
    """
    Base class for barcode detection results.

    Stores detected barcode, UMI, and quality scores for a single read.
    Implements the BarcodeResult protocol.
    """

    NOSEQ = "*"  # Sentinel for missing/undetected sequence

    def __init__(self, read_id: str, barcode: str = NOSEQ, UMI: str = NOSEQ,
                 BC_score: int = -1, UMI_good: bool = False, strand: str = "."):
        """
        Initialize barcode detection result.

        Args:
            read_id: Read identifier
            barcode: Detected barcode sequence (NOSEQ if not found)
            UMI: Detected UMI sequence (NOSEQ if not found)
            BC_score: Barcode alignment score
            UMI_good: Whether UMI passes quality filters
            strand: Detected strand ('+', '-', or '.')
        """
        self.read_id: str = read_id
        self._barcode: str = barcode if barcode else BarcodeDetectionResult.NOSEQ
        self._umi: str = UMI if UMI else BarcodeDetectionResult.NOSEQ
        self.BC_score: int = BC_score
        self.UMI_good: bool = UMI_good
        self.strand: str = strand

    # Primary getters (preferred interface)
    def get_barcode(self) -> str:
        """Get the detected barcode sequence."""
        return self._barcode

    def get_umi(self) -> str:
        """Get the detected UMI sequence."""
        return self._umi

    # Backward-compatible property accessors
    @property
    def barcode(self) -> str:
        """Barcode sequence. Prefer get_barcode() for new code."""
        return self._barcode

    @barcode.setter
    def barcode(self, value: str) -> None:
        self._barcode = value if value else BarcodeDetectionResult.NOSEQ

    @property
    def UMI(self) -> str:
        """UMI sequence. Prefer get_umi() for new code."""
        return self._umi

    @UMI.setter
    def UMI(self, value: str) -> None:
        self._umi = value if value else BarcodeDetectionResult.NOSEQ

    def is_valid(self) -> bool:
        """Check if a valid barcode was detected."""
        return self._barcode != BarcodeDetectionResult.NOSEQ

    def has_barcode(self) -> bool:
        """Check if a barcode was detected (alias for is_valid())."""
        return self._barcode != BarcodeDetectionResult.NOSEQ

    def has_umi(self) -> bool:
        """Check if a valid UMI was detected."""
        return self._umi != BarcodeDetectionResult.NOSEQ

    def update_coordinates(self, delta: int) -> None:
        """
        Shift all genomic coordinates by delta.

        Used when processing read subsequences.

        Args:
            delta: Amount to shift coordinates
        """
        pass

    def more_informative_than(self, that: 'BarcodeDetectionResult') -> bool:
        """
        Compare two results to determine which is more informative.

        Args:
            that: Another detection result

        Returns:
            True if this result is more informative

        Raises:
            NotImplementedError: Must be implemented by subclasses
        """
        raise NotImplementedError()

    def get_additional_attributes(self) -> List[str]:
        """
        Get list of detected additional features (primer, linker, etc.).

        Returns:
            List of detected feature names. Empty list for base class.
        """
        return []

    def set_strand(self, strand: str) -> None:
        """Set the detected strand."""
        self.strand = strand

    def __str__(self) -> str:
        """Format result as TSV line."""
        return "%s\t%s\t%s\t%d\t%s\t%s" % (self.read_id, self._barcode, self._umi,
                                           self.BC_score, self.UMI_good, self.strand)

    @staticmethod
    def header() -> str:
        """Static header for class-level access."""
        return "read_id\tbarcode\tUMI\tBC_score\tvalid_UMI\tstrand"


class LinkerBarcodeDetectionResult(BarcodeDetectionResult):
    """
    Detection result for platforms with double barcodes (e.g., Curio, Stereo-seq).

    Extends base result with positions of additional features:
    polyT tail, primer, and linker sequences.
    """

    def __init__(self, read_id: str, barcode: str = BarcodeDetectionResult.NOSEQ,
                 UMI: str = BarcodeDetectionResult.NOSEQ,
                 BC_score: int = -1, UMI_good: bool = False, strand: str = ".",
                 polyT: int = -1, primer: int = -1, linker_start: int = -1, linker_end: int = -1):
        """
        Initialize double barcode detection result.

        Args:
            read_id: Read identifier
            barcode: Detected barcode (concatenated if split by linker)
            UMI: Detected UMI sequence
            BC_score: Barcode alignment score
            UMI_good: Whether UMI passes quality filters
            strand: Detected strand
            polyT: Position of polyT tail start (-1 if not found)
            primer: Position of primer end (-1 if not found)
            linker_start: Position of linker start (-1 if not found)
            linker_end: Position of linker end (-1 if not found)
        """
        BarcodeDetectionResult.__init__(self, read_id, barcode, UMI, BC_score, UMI_good, strand)
        self.primer: int = primer
        self.linker_start: int = linker_start
        self.linker_end: int = linker_end
        self.polyT: int = polyT

    def update_coordinates(self, delta: int) -> None:
        self.primer = increase_if_valid(self.primer, delta)
        self.linker_start = increase_if_valid(self.linker_start, delta)
        self.linker_end = increase_if_valid(self.linker_end, delta)
        self.polyT = increase_if_valid(self.polyT, delta)

    def more_informative_than(self, that: 'LinkerBarcodeDetectionResult') -> bool:
        if self.BC_score != that.BC_score:
            return self.BC_score > that.BC_score
        if self.linker_start != that.linker_start:
            return self.linker_start > that.linker_start
        if self.primer != that.primer:
            return self.primer > that.primer
        return self.polyT > that.polyT

    def get_additional_attributes(self) -> List[str]:
        attr = []
        if self.polyT != -1:
            attr.append("PolyT detected")
        if self.primer != -1:
            attr.append("Primer detected")
        if self.linker_start != -1:
            attr.append("Linker detected")
        return attr

    def __str__(self) -> str:
        return (BarcodeDetectionResult.__str__(self) +
                "\t%d\t%d\t%d\t%d" % (self.polyT, self.primer, self.linker_start, self.linker_end))

    @staticmethod
    def header() -> str:
        """Static header for class-level access."""
        return BarcodeDetectionResult.header() + "\tpolyT_start\tprimer_end\tlinker_start\tlinker_end"


class TSOBarcodeDetectionResult(LinkerBarcodeDetectionResult):
    """Detection result for Stereo-seq with TSO detection."""

    def __init__(self, read_id: str, barcode: str = BarcodeDetectionResult.NOSEQ,
                 UMI: str = BarcodeDetectionResult.NOSEQ,
                 BC_score: int = -1, UMI_good: bool = False, strand: str = ".",
                 polyT: int = -1, primer: int = -1, linker_start: int = -1,
                 linker_end: int = -1, tso: int = -1):
        LinkerBarcodeDetectionResult.__init__(self, read_id, barcode, UMI, BC_score, UMI_good, strand,
                                              polyT, primer, linker_start, linker_end)
        self.tso5: int = tso

    def update_coordinates(self, delta: int) -> None:
        self.tso5 = increase_if_valid(self.tso5, delta)
        LinkerBarcodeDetectionResult.update_coordinates(self, delta)

    def __str__(self) -> str:
        return (LinkerBarcodeDetectionResult.__str__(self) +
                "\t%d" % self.tso5)

    def get_additional_attributes(self) -> List[str]:
        attr = []
        if self.polyT != -1:
            attr.append("PolyT detected")
        if self.primer != -1:
            attr.append("Primer detected")
        if self.linker_start != -1:
            attr.append("Linker detected")
        if self.tso5 != -1:
            attr.append("TSO detected")
        return attr

    @staticmethod
    def header() -> str:
        """Static header for class-level access."""
        return LinkerBarcodeDetectionResult.header() + "\tTSO5"


class TenXBarcodeDetectionResult(BarcodeDetectionResult):
    """Detection result for 10x Genomics platforms."""

    def __init__(self, read_id: str, barcode: str = BarcodeDetectionResult.NOSEQ,
                 UMI: str = BarcodeDetectionResult.NOSEQ,
                 BC_score: int = -1, UMI_good: bool = False, strand: str = ".",
                 polyT: int = -1, r1: int = -1):
        BarcodeDetectionResult.__init__(self, read_id, barcode, UMI, BC_score, UMI_good, strand)
        self.r1: int = r1
        self.polyT: int = polyT

    def update_coordinates(self, delta: int) -> None:
        self.r1 = increase_if_valid(self.r1, delta)
        self.polyT = increase_if_valid(self.polyT, delta)

    def more_informative_than(self, that: 'TenXBarcodeDetectionResult') -> bool:
        if self.polyT != that.polyT:
            return self.polyT > that.polyT
        if self.r1 != that.r1:
            return self.r1 > that.r1
        return self.BC_score > that.BC_score

    def get_additional_attributes(self) -> List[str]:
        attr = []
        if self.polyT != -1:
            attr.append("PolyT detected")
        if self.r1 != -1:
            attr.append("R1 detected")
        return attr

    def __str__(self) -> str:
        return (BarcodeDetectionResult.__str__(self) +
                "\t%d\t%d" % (self.polyT, self.r1))

    @staticmethod
    def header() -> str:
        """Static header for class-level access."""
        return BarcodeDetectionResult.header() + "\tpolyT_start\tR1_end"


class SplittingBarcodeDetectionResult:
    """Result container for read splitting modes (multiple barcodes per read)."""

    NOSEQ = BarcodeDetectionResult.NOSEQ  # For consistency with protocol

    def __init__(self, read_id: str):
        self.read_id: str = read_id
        self.strand: str = "."  # For protocol compatibility
        self.detected_patterns: List[TSOBarcodeDetectionResult] = []

    def append(self, barcode_detection_result: TSOBarcodeDetectionResult) -> None:
        self.detected_patterns.append(barcode_detection_result)

    def empty(self) -> bool:
        return not self.detected_patterns

    def filter(self) -> None:
        if not self.detected_patterns:
            return
        barcoded_results = []
        for r in self.detected_patterns:
            if r.is_valid():
                barcoded_results.append(r)

        if not barcoded_results:
            self.detected_patterns = [self.detected_patterns[0]]
        else:
            self.detected_patterns = barcoded_results

    def get_barcode(self) -> str:
        """Get barcode from first valid pattern, or NOSEQ if none."""
        for r in self.detected_patterns:
            if r.is_valid():
                return r.get_barcode()
        return self.NOSEQ

    def get_umi(self) -> str:
        """Get UMI from first valid pattern, or NOSEQ if none."""
        for r in self.detected_patterns:
            if r.is_valid():
                return r.get_umi()
        return self.NOSEQ

    def is_valid(self) -> bool:
        """Check if any pattern has a valid barcode."""
        return any(r.is_valid() for r in self.detected_patterns)

    def has_barcode(self) -> bool:
        """Check if any pattern has a valid barcode."""
        return any(r.has_barcode() for r in self.detected_patterns)

    def has_umi(self) -> bool:
        """Check if any pattern has a valid UMI."""
        return any(r.has_umi() for r in self.detected_patterns)

    def set_strand(self, strand: str) -> None:
        """Set strand for all patterns."""
        self.strand = strand
        for r in self.detected_patterns:
            r.set_strand(strand)

    def update_coordinates(self, delta: int) -> None:
        """Update coordinates for all patterns."""
        for r in self.detected_patterns:
            r.update_coordinates(delta)

    def more_informative_than(self, other: 'SplittingBarcodeDetectionResult') -> bool:
        """Compare by number of valid patterns."""
        self_valid = sum(1 for r in self.detected_patterns if r.is_valid())
        other_valid = sum(1 for r in other.detected_patterns if r.is_valid())
        return self_valid > other_valid

    def get_additional_attributes(self) -> List[str]:
        """Get combined attributes from all patterns."""
        attrs = set()
        for r in self.detected_patterns:
            attrs.update(r.get_additional_attributes())
        return list(attrs)

    def __str__(self) -> str:
        """Format all patterns as TSV lines."""
        return "\n".join(str(r) for r in self.detected_patterns)

    @staticmethod
    def header() -> str:
        """Get TSV header for result output."""
        return TSOBarcodeDetectionResult.header()


class ReadStats:
    """
    Statistics tracker for barcode detection results.

    Accumulates counts of processed reads, detected barcodes, valid UMIs,
    and platform-specific features (primers, linkers, polyT tails, etc.).
    """

    def __init__(self):
        """Initialize empty statistics."""
        self.read_count: int = 0
        self.bc_count: int = 0
        self.umi_count: int = 0
        self.additional_attributes_counts: Dict[str, int] = defaultdict(int)

    def add_read(self, barcode_detection_result) -> None:
        """
        Add a read result to statistics.

        Args:
            barcode_detection_result: Detection result to accumulate (implements BarcodeResult protocol)
        """
        self.read_count += 1
        # Count detected features (primer, linker, etc.)
        for a in barcode_detection_result.get_additional_attributes():
            self.additional_attributes_counts[a] += 1
        # Count valid barcode
        if barcode_detection_result.has_barcode():
            self.bc_count += 1
        # Count valid UMI
        if barcode_detection_result.has_umi():
            self.umi_count += 1

    def add_custom_stats(self, stat_name: str, val: int) -> None:
        """
        Add custom statistic value.

        Args:
            stat_name: Name of statistic
            val: Count to add
        """
        self.additional_attributes_counts[stat_name] += val

    def __str__(self) -> str:
        """Format statistics as human-readable string."""
        human_readable_str = ("Total reads\t%d\nBarcode detected\t%d\nReliable UMI\t%d\n" %
                              (self.read_count, self.bc_count, self.umi_count))
        for a in self.additional_attributes_counts:
            human_readable_str += "%s\t%d\n" % (a, self.additional_attributes_counts[a])
        return human_readable_str

    def __iter__(self) -> Iterable[str]:
        """Iterate over statistics as formatted strings."""
        yield "Total reads: %d" % self.read_count
        yield "Barcode detected: %d" % self.bc_count
        yield "Reliable UMI: %d" % self.umi_count
        for a in self.additional_attributes_counts:
            yield "%s: %d" % (a, self.additional_attributes_counts[a])



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
from src.barcode_calling.callers import (
    BarcodeDetectionResult,
    LinkerBarcodeDetectionResult,
    TSOBarcodeDetectionResult,
    TenXBarcodeDetectionResult,
    SplittingBarcodeDetectionResult,
    ReadStats,
    increase_if_valid
)


class TestIncreaseIfValid:
    """Test coordinate increment utility function."""

    def test_increment_valid_positive(self):
        """Test incrementing valid positive value."""
        assert increase_if_valid(10, 5) == 15

    def test_increment_valid_zero(self):
        """Test incrementing zero (treated as invalid)."""
        assert increase_if_valid(0, 5) == 0

    def test_increment_invalid_negative(self):
        """Test incrementing -1 (invalid marker)."""
        assert increase_if_valid(-1, 5) == -1

    def test_increment_none(self):
        """Test incrementing None."""
        assert increase_if_valid(None, 5) is None


class TestBarcodeDetectionResult:
    """Test base barcode detection result class."""

    def test_init_default(self):
        """Test initialization with defaults."""
        result = BarcodeDetectionResult("read_001")

        assert result.read_id == "read_001"
        assert result.barcode == BarcodeDetectionResult.NOSEQ
        assert result.UMI == BarcodeDetectionResult.NOSEQ
        assert result.BC_score == -1
        assert result.UMI_good is False
        assert result.strand == "."

    def test_init_with_values(self):
        """Test initialization with all values."""
        result = BarcodeDetectionResult(
            read_id="read_001",
            barcode="ACTGACTG",
            UMI="GGGGGGGG",
            BC_score=16,
            UMI_good=True,
            strand="+"
        )

        assert result.read_id == "read_001"
        assert result.barcode == "ACTGACTG"
        assert result.UMI == "GGGGGGGG"
        assert result.BC_score == 16
        assert result.UMI_good is True
        assert result.strand == "+"

    def test_is_valid_with_barcode(self):
        """Test validity check with detected barcode."""
        result = BarcodeDetectionResult("read_001", barcode="ACTG")
        assert result.is_valid() is True

    def test_is_valid_without_barcode(self):
        """Test validity check without barcode."""
        result = BarcodeDetectionResult("read_001")
        assert result.is_valid() is False

    def test_set_strand(self):
        """Test setting strand."""
        result = BarcodeDetectionResult("read_001")
        result.set_strand("+")
        assert result.strand == "+"

    def test_str_format(self):
        """Test string formatting."""
        result = BarcodeDetectionResult(
            "read_001", "ACTG", "GGGG", 15, True, "+"
        )
        output = str(result)

        assert "read_001" in output
        assert "ACTG" in output
        assert "GGGG" in output
        assert "15" in output
        assert "True" in output
        assert "+" in output

    def test_header(self):
        """Test TSV header."""
        # header() is a static method
        header = BarcodeDetectionResult.header()
        assert "read_id" in header
        assert "barcode" in header
        assert "UMI" in header
        assert "BC_score" in header


class TestCurioBarcodeDetectionResult:
    """Test double barcode detection result class."""

    def test_init_default(self):
        """Test initialization with defaults."""
        result = LinkerBarcodeDetectionResult("read_001")

        assert result.read_id == "read_001"
        assert result.polyT == -1
        assert result.primer == -1
        assert result.linker_start == -1
        assert result.linker_end == -1

    def test_init_with_positions(self):
        """Test initialization with all positions."""
        result = LinkerBarcodeDetectionResult(
            read_id="read_001",
            barcode="ACTGACTG",
            UMI="GGGG",
            BC_score=14,
            UMI_good=True,
            strand="+",
            polyT=100,
            primer=50,
            linker_start=60,
            linker_end=75
        )

        assert result.polyT == 100
        assert result.primer == 50
        assert result.linker_start == 60
        assert result.linker_end == 75

    def test_update_coordinates(self):
        """Test coordinate shifting."""
        result = LinkerBarcodeDetectionResult(
            "read_001",
            polyT=100,
            primer=50,
            linker_start=60,
            linker_end=75
        )

        result.update_coordinates(10)

        assert result.polyT == 110
        assert result.primer == 60
        assert result.linker_start == 70
        assert result.linker_end == 85

    def test_update_coordinates_invalid(self):
        """Test coordinate shifting with invalid values."""
        result = LinkerBarcodeDetectionResult("read_001")
        result.update_coordinates(10)

        # Invalid values (-1) should remain unchanged
        assert result.polyT == -1
        assert result.primer == -1

    def test_more_informative_than_by_score(self):
        """Test comparison by barcode score."""
        result1 = LinkerBarcodeDetectionResult("read_001", BC_score=16)
        result2 = LinkerBarcodeDetectionResult("read_001", BC_score=14)

        assert result1.more_informative_than(result2) is True
        assert result2.more_informative_than(result1) is False

    def test_more_informative_than_by_linker(self):
        """Test comparison by linker position (tie-breaker)."""
        result1 = LinkerBarcodeDetectionResult(
            "read_001", BC_score=16, linker_start=80
        )
        result2 = LinkerBarcodeDetectionResult(
            "read_001", BC_score=16, linker_start=60
        )

        # Higher linker position is more informative
        assert result1.more_informative_than(result2) is True

    def test_get_additional_attributes_all(self):
        """Test attribute detection with all features."""
        result = LinkerBarcodeDetectionResult(
            "read_001",
            polyT=100,
            primer=50,
            linker_start=60
        )

        attrs = result.get_additional_attributes()

        assert "PolyT detected" in attrs
        assert "Primer detected" in attrs
        assert "Linker detected" in attrs

    def test_get_additional_attributes_partial(self):
        """Test attribute detection with some features."""
        result = LinkerBarcodeDetectionResult(
            "read_001",
            polyT=100,
            # No primer or linker
        )

        attrs = result.get_additional_attributes()

        assert "PolyT detected" in attrs
        assert "Primer detected" not in attrs
        assert "Linker detected" not in attrs

    def test_str_format(self):
        """Test string formatting includes positions."""
        result = LinkerBarcodeDetectionResult(
            "read_001", "ACTG", "GGGG", 16, True, "+",
            polyT=100, primer=50, linker_start=60, linker_end=75
        )
        output = str(result)

        assert "100" in output  # polyT
        assert "50" in output   # primer
        assert "60" in output   # linker_start
        assert "75" in output   # linker_end


class TestStereoBarcodeDetectionResult:
    """Test Stereo-seq detection result class."""

    def test_init_with_tso(self):
        """Test initialization with TSO position."""
        result = TSOBarcodeDetectionResult(
            "read_001",
            polyT=100,
            tso=150
        )

        assert result.tso5 == 150

    def test_update_coordinates_includes_tso(self):
        """Test coordinate shifting includes TSO."""
        result = TSOBarcodeDetectionResult(
            "read_001",
            polyT=100,
            tso=150
        )

        result.update_coordinates(10)

        assert result.tso5 == 160
        assert result.polyT == 110

    def test_get_additional_attributes_with_tso(self):
        """Test attribute detection includes TSO."""
        result = TSOBarcodeDetectionResult(
            "read_001",
            polyT=100,
            tso=150
        )

        attrs = result.get_additional_attributes()

        assert "PolyT detected" in attrs
        assert "TSO detected" in attrs


class TestTenXBarcodeDetectionResult:
    """Test 10x Genomics detection result class."""

    def test_init_with_r1(self):
        """Test initialization with R1 position."""
        result = TenXBarcodeDetectionResult(
            "read_001",
            polyT=100,
            r1=20
        )

        assert result.r1 == 20
        assert result.polyT == 100

    def test_update_coordinates(self):
        """Test coordinate shifting."""
        result = TenXBarcodeDetectionResult(
            "read_001",
            polyT=100,
            r1=20
        )

        result.update_coordinates(10)

        assert result.r1 == 30
        assert result.polyT == 110

    def test_more_informative_than_by_polyt(self):
        """Test comparison prioritizes polyT."""
        result1 = TenXBarcodeDetectionResult("read_001", polyT=100)
        result2 = TenXBarcodeDetectionResult("read_001", polyT=80)

        assert result1.more_informative_than(result2) is True

    def test_more_informative_than_by_r1(self):
        """Test comparison uses R1 as tie-breaker."""
        result1 = TenXBarcodeDetectionResult("read_001", polyT=100, r1=30)
        result2 = TenXBarcodeDetectionResult("read_001", polyT=100, r1=20)

        assert result1.more_informative_than(result2) is True

    def test_get_additional_attributes(self):
        """Test attribute detection."""
        result = TenXBarcodeDetectionResult(
            "read_001",
            polyT=100,
            r1=20
        )

        attrs = result.get_additional_attributes()

        assert "PolyT detected" in attrs
        assert "R1 detected" in attrs


class TestSplittingBarcodeDetectionResult:
    """Test splitting barcode detection result class."""

    def test_init(self):
        """Test initialization."""
        result = SplittingBarcodeDetectionResult("read_001")

        assert result.read_id == "read_001"
        assert result.detected_patterns == []

    def test_append(self):
        """Test appending detection patterns."""
        result = SplittingBarcodeDetectionResult("read_001")

        pattern1 = TSOBarcodeDetectionResult("read_001", barcode="ACTG")
        pattern2 = TSOBarcodeDetectionResult("read_001", barcode="TGCA")

        result.append(pattern1)
        result.append(pattern2)

        assert len(result.detected_patterns) == 2

    def test_empty_true(self):
        """Test empty detection."""
        result = SplittingBarcodeDetectionResult("read_001")
        assert result.empty() is True

    def test_empty_false(self):
        """Test non-empty detection."""
        result = SplittingBarcodeDetectionResult("read_001")
        pattern = TSOBarcodeDetectionResult("read_001", barcode="ACTG")
        result.append(pattern)

        assert result.empty() is False

    def test_filter_keeps_barcoded(self):
        """Test filter keeps results with barcodes."""
        result = SplittingBarcodeDetectionResult("read_001")

        barcoded = TSOBarcodeDetectionResult("read_001", barcode="ACTG")
        unbarcoded = TSOBarcodeDetectionResult("read_001")  # No barcode

        result.append(barcoded)
        result.append(unbarcoded)
        result.filter()

        # Should keep only barcoded result
        assert len(result.detected_patterns) == 1
        assert result.detected_patterns[0].barcode == "ACTG"

    def test_filter_keeps_first_if_none_barcoded(self):
        """Test filter keeps first result if none have barcodes."""
        result = SplittingBarcodeDetectionResult("read_001")

        unbarcoded1 = TSOBarcodeDetectionResult("read_001")
        unbarcoded2 = TSOBarcodeDetectionResult("read_001")

        result.append(unbarcoded1)
        result.append(unbarcoded2)
        result.filter()

        # Should keep only first result
        assert len(result.detected_patterns) == 1


class TestReadStats:
    """Test statistics tracker class."""

    def test_init(self):
        """Test initialization."""
        stats = ReadStats()

        assert stats.read_count == 0
        assert stats.bc_count == 0
        assert stats.umi_count == 0
        assert len(stats.additional_attributes_counts) == 0

    def test_add_read_with_barcode(self):
        """Test adding read with valid barcode."""
        stats = ReadStats()
        result = LinkerBarcodeDetectionResult(
            "read_001",
            barcode="ACTG",
            UMI="GGGGGGGG",  # Need actual UMI for has_umi() to return True
            UMI_good=True,
            polyT=100
        )

        stats.add_read(result)

        assert stats.read_count == 1
        assert stats.bc_count == 1
        assert stats.umi_count == 1
        assert stats.additional_attributes_counts["PolyT detected"] == 1

    def test_add_read_without_barcode(self):
        """Test adding read without barcode."""
        stats = ReadStats()
        result = LinkerBarcodeDetectionResult("read_001")  # No barcode

        stats.add_read(result)

        assert stats.read_count == 1
        assert stats.bc_count == 0
        assert stats.umi_count == 0

    def test_add_multiple_reads(self):
        """Test adding multiple reads."""
        stats = ReadStats()

        result1 = LinkerBarcodeDetectionResult("read_001", barcode="ACTG", polyT=100)
        result2 = LinkerBarcodeDetectionResult("read_002", barcode="TGCA", primer=50)
        result3 = LinkerBarcodeDetectionResult("read_003")  # No barcode

        stats.add_read(result1)
        stats.add_read(result2)
        stats.add_read(result3)

        assert stats.read_count == 3
        assert stats.bc_count == 2
        assert stats.additional_attributes_counts["PolyT detected"] == 1
        assert stats.additional_attributes_counts["Primer detected"] == 1

    def test_add_custom_stats(self):
        """Test adding custom statistics."""
        stats = ReadStats()

        stats.add_custom_stats("Custom feature", 5)
        stats.add_custom_stats("Custom feature", 3)

        assert stats.additional_attributes_counts["Custom feature"] == 8

    def test_str_format(self):
        """Test string formatting."""
        stats = ReadStats()
        # Need actual UMI for has_umi() to return True
        result = LinkerBarcodeDetectionResult("read_001", barcode="ACTG", UMI="GGGGGGGG", UMI_good=True)
        stats.add_read(result)

        output = str(stats)

        assert "Total reads\t1" in output
        assert "Barcode detected\t1" in output
        assert "Reliable UMI\t1" in output

    def test_iter(self):
        """Test iteration over statistics."""
        stats = ReadStats()
        result = LinkerBarcodeDetectionResult("read_001", barcode="ACTG", polyT=100)
        stats.add_read(result)

        lines = list(stats)

        assert "Total reads: 1" in lines
        assert "Barcode detected: 1" in lines
        assert "PolyT detected: 1" in lines


if __name__ == '__main__':
    pytest.main([__file__, '-v'])
